@@ -1,5 +1,5 @@
 //
-// Copyright (C) 2010-2019 Codership Oy <info@codership.com>
+// Copyright (C) 2010-2020 Codership Oy <info@codership.com>
 //
 
 #include "replicator_smm.hpp"
@@ -58,30 +58,10 @@ ReplicatorSMM::state_transfer_required(const wsrep_view_info_t& view_info,
             wsrep_seqno_t const group_seqno(view_info.state_id.seqno);
             wsrep_seqno_t const local_seqno(last_committed());
 
-            if (state_() >= S_JOINING) /* See #442 - S_JOINING should be
-                                          a valid state here */
-            {
-                if (str_proto_ver >= 3)
-                    return (local_seqno + 1 < group_seqno); // this CC will add 1
-                else
-                    return (local_seqno < group_seqno);
-            }
+            if (str_proto_ver >= 3)
+                return (local_seqno + 1 < group_seqno); // this CC will add 1
             else
-            {
-                if ((str_proto_ver >= 3 && local_seqno >= group_seqno) ||
-                    (str_proto_ver <  3 && local_seqno >  group_seqno))
-                {
-                    close();
-                    gu_throw_fatal
-                        << "Local state seqno (" << local_seqno
-                        << ") is greater than group seqno (" <<group_seqno
-                        << "): states diverged. Aborting to avoid potential "
-                        << "data loss. Remove '" << state_file_
-                        << "' file and restart if you wish to continue.";
-                }
-
-                return (local_seqno != group_seqno);
-            }
+                return (local_seqno < group_seqno);
         }
 
         return true;
@@ -240,7 +220,6 @@ StateRequest_v1::StateRequest_v1 (const void* str, ssize_t str_len)
 {
     if (sst_offset() + 2*sizeof(uint32_t) > size_t(len_))
     {
-        assert(0);
         gu_throw_error (EINVAL) << "State transfer request is too short: "
                                 << len_ << ", must be at least: "
                                 << (sst_offset() + 2*sizeof(uint32_t));
@@ -248,7 +227,6 @@ StateRequest_v1::StateRequest_v1 (const void* str, ssize_t str_len)
 
     if (strncmp (req_, MAGIC.c_str(), MAGIC.length()))
     {
-        assert(0);
         gu_throw_error (EINVAL) << "Wrong magic signature in state request v1.";
     }
 
@@ -382,6 +360,44 @@ ReplicatorSMM::donate_sst(void* const         recv_ctx,
     return ret;
 }
 
+struct slg
+{
+    gcache::GCache& gcache_;
+    bool            unlock_;
+
+    slg(gcache::GCache& cache) : gcache_(cache), unlock_(false){}
+    ~slg() { if (unlock_) gcache_.seqno_unlock(); }
+};
+
+static wsrep_seqno_t run_ist_senders(ist::AsyncSenderMap& ist_senders,
+                                     const gu::Config&    config,
+                                     const std::string&   peer,
+                                     wsrep_seqno_t const  preload_start,
+                                     wsrep_seqno_t const  cc_seqno,
+                                     wsrep_seqno_t const  cc_lowest,
+                                     int const            proto_ver,
+                                     slg&                 seqno_lock_guard,
+                                     wsrep_seqno_t const  rcode)
+{
+    try
+    {
+        ist_senders.run(config,
+                        peer,
+                        preload_start,
+                        cc_seqno,
+                        cc_lowest,
+                        proto_ver);
+        // seqno will be unlocked when sender exists
+        seqno_lock_guard.unlock_ = false;
+        return rcode;
+    }
+    catch (gu::Exception& e)
+    {
+        log_warn << "IST failed: " << e.what();
+        return -e.get_errno();
+    }
+}
+
 void ReplicatorSMM::process_state_req(void*       recv_ctx,
                                       const void* req,
                                       size_t      req_size,
@@ -392,7 +408,8 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
     assert(seqno_l > -1);
     assert(req != 0);
 
-    StateRequest* const streq(read_state_request(req, req_size));
+    StateRequest* const streq(read_state_request(
+                                  gcache_.get_ro_plaintext(req), req_size));
     // Guess correct STR protocol version. Here we assume that the
     // replicator protocol version didn't change between sending
     // and receiving STR message. Unfortunately the protocol version
@@ -426,73 +443,77 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
     wsrep_seqno_t rcode (0);
     bool join_now = true;
 
-    if (streq->ist_len())
+    if (not skip_sst)
     {
-        IST_request istr;
-        get_ist_request(streq, &istr);
+        slg seqno_lock_guard(gcache_);
 
-        if (istr.uuid() == state_uuid_ && istr.last_applied() >= 0)
+        if (streq->ist_len())
         {
-            log_info << "IST request: " << istr;
+            IST_request istr;
+            get_ist_request(streq, &istr);
 
-            try
+            if (istr.uuid() == state_uuid_ && istr.last_applied() >= 0)
             {
-                gcache_.seqno_lock(istr.last_applied() + 1);
-            }
-            catch(gu::NotFound& nf)
-            {
-                log_info << "IST first seqno " << istr.last_applied() + 1
-                         << " not found from cache, falling back to SST";
-                // @todo: close IST channel explicitly
-                goto full_sst;
-            }
+                log_info << "IST request: " << istr;
 
-            if (streq->sst_len()) // if joiner is waiting for SST, notify it
-            {
-                wsrep_gtid_t const state_id =
-                    { istr.uuid(), istr.last_applied() };
-
-                rcode = donate_sst(recv_ctx, *streq, state_id, true);
-
-                // we will join in sst_sent.
-                join_now = false;
-            }
-
-            if (rcode >= 0)
-            {
                 wsrep_seqno_t const first
                     ((str_proto_ver < 3 || cc_lowest_trx_seqno_ == 0) ?
-                     istr.last_applied() + 1 :
-                     std::min(cc_lowest_trx_seqno_, istr.last_applied()+1));
+                    istr.last_applied() + 1 :
+                    std::min(cc_lowest_trx_seqno_, istr.last_applied()+1));
+
                 try
                 {
-                    ist_senders_.run(config_,
-                                     istr.peer(),
-                                     first,
-                                     cc_seqno_,
-                                     cc_lowest_trx_seqno_,
-                                     /* Historically IST messages versioned
-                                      * with the global replicator protocol.
-                                      * Need to keep it that way for backward
-                                      * compatibility */
-                                     protocol_version_);
+                    gcache_.seqno_lock(first);
+                    seqno_lock_guard.unlock_ = true;
                 }
-                catch (gu::Exception& e)
+                catch(gu::NotFound& nf)
                 {
-                    log_error << "IST failed: " << e.what();
-                    rcode = -e.get_errno();
+                    log_info << "IST first seqno " << istr.last_applied() + 1
+                             << " not found from cache, falling back to SST";
+                    // @todo: close IST channel explicitly
+                    goto full_sst;
                 }
-            }
-            else
-            {
-                log_error << "Failed to bypass SST";
-            }
 
-            goto out;
+                if (streq->sst_len()) // if joiner is waiting for SST, notify it
+                {
+                    wsrep_gtid_t const state_id =
+                        { istr.uuid(), istr.last_applied() };
+
+                    gu_trace(rcode = donate_sst(recv_ctx, *streq, state_id, true));
+
+                    // we will join in sst_sent.
+                    join_now = false;
+                }
+
+                if (rcode >= 0)
+                {
+                    rcode = run_ist_senders(ist_senders_,
+                                            config_,
+                                            istr.peer(),
+                                            first,
+                                            cc_seqno_,
+                                            cc_lowest_trx_seqno_,
+                        /* Historically IST messages are versioned
+                         * with the global replicator protocol.
+                         * Need to keep it that way for backward
+                         * compatibility */
+                                            protocol_version_,
+                                            seqno_lock_guard,
+                                            rcode);
+                }
+                else
+                {
+                    log_error << "Failed to bypass SST";
+                }
+
+                goto out;
+            }
         }
 
     full_sst:
 
+        assert(!seqno_lock_guard.unlock_);
+
         if (cert_.nbo_size() > 0)
         {
             log_warn << "Non-blocking operation in progress, cannot donate SST";
@@ -536,6 +557,7 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
                         }
 
                         gcache_.seqno_lock(preload_start);
+                        seqno_lock_guard.unlock_ = true;
                     }
                     catch (gu::NotFound& nf)
                     {
@@ -553,16 +575,20 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
                     IST_request istr;
                     get_ist_request(streq, &istr);
                     // Send trxs to rebuild cert index.
-                    ist_senders_.run(config_,
-                                     istr.peer(),
-                                     preload_start,
-                                     cc_seqno_,
-                                     preload_start,
-                                     /* Historically IST messages are versioned
-                                      * with the global replicator protocol.
-                                      * Need to keep it that way for backward
-                                      * compatibility */
-                                     protocol_version_);
+                    rcode = run_ist_senders(ist_senders_,
+                                            config_,
+                                            istr.peer(),
+                                            preload_start,
+                                            cc_seqno_,
+                                            preload_start,
+                        /* Historically IST messages are versioned
+                         * with the global replicator protocol.
+                         * Need to keep it that way for backward
+                         * compatibility */
+                                            protocol_version_,
+                                            seqno_lock_guard,
+                                            rcode);
+                    if (rcode < 0) goto out;
                 }
                 else /* streq->version() == 0 */
                 {
@@ -571,12 +597,9 @@ void ReplicatorSMM::process_state_req(void*       recv_ctx,
                 }
             }
 
-            if (not skip_sst)
-            {
-                rcode = donate_sst(recv_ctx, *streq, state_id, false);
-                // we will join in sst_sent.
-                join_now = false;
-            }
+            rcode = donate_sst(recv_ctx, *streq, state_id, false);
+            // we will join in sst_sent.
+            join_now = false;
         }
         else
         {
@@ -870,11 +893,13 @@ ReplicatorSMM::request_state_transfer (void* recv_ctx,
 
     st_.mark_unsafe();
 
+    GU_DBUG_SYNC_WAIT("before_send_state_request");
     send_state_request(req, str_proto_ver);
 
     state_.shift_to(S_JOINING);
     sst_state_ = SST_WAIT;
     sst_seqno_ = WSREP_SEQNO_UNDEFINED;
+    GU_DBUG_SYNC_WAIT("after_shift_to_joining");
 
     /* There are two places where we may need to adjust GCache.
      * This is the first one, which we can do while waiting for SST to complete.
@@ -937,13 +962,25 @@ ReplicatorSMM::request_state_transfer (void* recv_ctx,
 
             update_state_uuid (sst_uuid_);
 
-            if (str_proto_ver < 3)
+            if (group_proto_ver < PROTO_VER_GALERA_3_MAX)
             {
-                // all IST events will bypass certification
+                log_error << "Rolling upgrade from group protocol version "
+                          << "earlier than "
+                          << PROTO_VER_GALERA_3_MAX
+                          << " is not supported. Please upgrade "
+                          << "Galera library to latest in Galera 3 series on "
+                          << "all of the nodes in the cluster before "
+                          << "continuing.";
+                abort();
+            }
+            else if (group_proto_ver == PROTO_VER_GALERA_3_MAX)
+            {
+                // Rolling upgrade from Galera 3 PROTO_VER_GALERA_3_MAX.
                 gu::GTID const cert_position
                     (sst_uuid_, std::max(cc_seqno, sst_seqno_));
-                cert_.assign_initial_position(cert_position,
-                                              trx_params_.version_);
+                cert_.assign_initial_position(
+                    cert_position,
+                    std::get<0>(get_trx_protocol_versions(group_proto_ver)));
                 // with higher versions this happens in cert index preload
             }
 
@@ -957,7 +994,7 @@ ReplicatorSMM::request_state_transfer (void* recv_ctx,
             }
 
             log_info << "Installed new state from SST: " << state_uuid_ << ":"
-                      << sst_seqno_;
+                     << sst_seqno_;
         }
     }
     else
@@ -970,17 +1007,23 @@ ReplicatorSMM::request_state_transfer (void* recv_ctx,
     {
         if (sst_req_len != 0 && !sst_is_trivial(sst_req, sst_req_len))
         {
-            st_.mark_uncorrupt(sst_uuid_, sst_seqno_);
+            // Note: not storing sst seqno in state file to avoid
+            // recovering to incorrect state if the node is
+            // killed during IST.
+            st_.mark_uncorrupt(sst_uuid_, WSREP_SEQNO_UNDEFINED);
         }
         else
         {
             log_fatal << "Application state is corrupt and cannot "
-                      << "be recorvered. Restart required.";
+                      << "be recovered. Restart required.";
             abort();
         }
     }
     else
     {
+        // Clear seqno from state file. Otherwise if node gets killed
+        // during IST, it may recover to incorrect position.
+        st_.set(state_uuid_, WSREP_SEQNO_UNDEFINED, safe_to_bootstrap_);
         st_.mark_safe();
     }
 
@@ -1029,7 +1072,7 @@ ReplicatorSMM::request_state_transfer (void* recv_ctx,
                 // Note: apply_monitor_ must be drained to avoid race between
                 // IST appliers and GCS appliers, GCS action source may
                 // provide actions that have already been applied via IST.
-                log_info << "Draining apply monitors after IST upto "
+                log_info << "Draining apply monitors after IST up to "
                          << sst_seqno_;
                 apply_monitor_.drain(sst_seqno_);
                 set_initial_position(group_uuid, sst_seqno_);
@@ -1098,7 +1141,15 @@ void ReplicatorSMM::process_IST_writeset(void* recv_ctx,
         assert(ts.is_dummy());
     }
 
-    gu_trace(apply_trx(recv_ctx, ts));
+    try
+    {
+        apply_trx(recv_ctx, ts);
+    }
+    catch (...)
+    {
+        st_.mark_corrupt();
+        throw;
+    }
     GU_DBUG_SYNC_WAIT("recv_IST_after_apply_trx");
 
     if (gu_unlikely
@@ -1185,7 +1236,8 @@ void ReplicatorSMM::recv_IST(void* recv_ctx)
 
         log_fatal << os.str();
 
-        mark_corrupt_and_close();
+        gu::Lock lock(closing_mutex_);
+        start_closing();
     }
 }
 
@@ -1389,14 +1441,14 @@ void galera::ReplicatorSMM::process_ist_conf_change(const gcs_act_cchange& conf)
     ist_event_queue_.push_back(view_info);
 }
 
-void ReplicatorSMM::ist_cc(const gcs_action& act, bool must_apply,
-                           bool preload)
+void ReplicatorSMM::ist_cc(const gcs_act_cchange& conf,
+                           const gcs_action&      act,
+                           bool const             must_apply,
+                           bool const             preload)
 {
     assert(GCS_ACT_CCHANGE == act.type);
     assert(act.seqno_g > 0);
 
-    gcs_act_cchange const conf(act.buf, act.size);
-
     assert(conf.conf_id >= 0); // Primary configuration
     assert(conf.seqno == act.seqno_g);
 
@@ -1431,7 +1483,7 @@ void ReplicatorSMM::ist_cc(const gcs_action& act, bool must_apply,
             establish_protocol_versions(conf.repl_proto_ver);
             cert_.adjust_position(*view_info, gu::GTID(conf.uuid, conf.seqno),
                                   trx_params_.version_);
-            // record CC releated state seqnos, needed for IST on DONOR
+            // record CC related state seqnos, needed for IST on DONOR
             record_cc_seqnos(conf.seqno, "preload");
             ::free(view_info);
         }
