@@ -66,7 +66,11 @@ from_string(const std::string& s,
 
     try
     {
-        if ((iss >> f >> ret).fail()) throw NotFound();
+        iss >> f >> ret;
+        if (iss.fail() || not iss.eof())
+        {
+            throw NotFound();
+        }
     }
     catch (gu::Exception& e)
     {
@@ -93,7 +97,11 @@ from_string<void*>(const std::string& s,
     std::istringstream iss(s);
     void*              ret;
 
-    if ((iss >> std::hex >> ret).fail()) throw NotFound();
+    iss >> std::hex >> ret;
+    if (iss.fail() || not iss.eof())
+    {
+        throw NotFound();
+    }
 
     return ret;
 }
@@ -204,13 +212,14 @@ std::ostream& operator << (std::ostream& os, const PrintBase<base, T>& b)
 /*! template to do arithmetics on void and byte pointers, compiler will
  *  catch anything else.
  * @return input type */
-template <typename T>
+template <typename T, typename PtrOffsetType>
 inline T*
-ptr_offset(T* ptr, ptrdiff_t i) { return static_cast<byte_t*>(ptr) + i; }
+ptr_offset(T* ptr, PtrOffsetType i) { return static_cast<byte_t*>(ptr) + i; }
 
-template <typename T>
+template <typename T, typename PtrOffsetType>
 inline const T*
-ptr_offset(const T* ptr, ptrdiff_t i) { return static_cast<const byte_t*>(ptr)+i; }
+ptr_offset(const T* ptr, PtrOffsetType i)
+    { return static_cast<const byte_t*>(ptr)+i; }
 
 } // namespace gu
 
